/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*  Fluent Bit
 *  ==========
 *  Copyright (C) 2015-2017 Treasure Data Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#include <fluent-bit/flb_info.h>
#include <fluent-bit/flb_mem.h>

#include "es.h"
#include "es_conf.h"

static int bool_value(char *v)
{
    if (strcasecmp(v, "true") == 0) {
        return FLB_TRUE;
    }
    else if (strcasecmp(v, "on") == 0) {
        return FLB_TRUE;
    }
    else if (strcasecmp(v, "yes") == 0) {
        return FLB_TRUE;
    }

    return FLB_FALSE;
}

struct flb_elasticsearch *flb_es_conf_create(struct flb_output_instance *ins,
                                             struct flb_config *config)
{

    int io_type;
    char *tmp;
    struct flb_uri *uri = ins->host.uri;
    struct flb_uri_field *f_index = NULL;
    struct flb_uri_field *f_type = NULL;
    struct flb_upstream *upstream;
    struct flb_elasticsearch *ctx;

    /* Allocate context */
    ctx = flb_calloc(1, sizeof(struct flb_elasticsearch));
    if (!ctx) {
        flb_errno();
        return NULL;
    }

    if (uri) {
        if (uri->count >= 2) {
            f_index = flb_uri_get(uri, 0);
            f_type  = flb_uri_get(uri, 1);
        }
    }

    /* Get network configuration */
    if (!ins->host.name) {
        ins->host.name = flb_strdup("127.0.0.1");
    }

    if (ins->host.port == 0) {
        ins->host.port = 9200;
    }

    /* use TLS ? */
    if (ins->use_tls == FLB_TRUE) {
        io_type = FLB_IO_TLS;
    }
    else {
        io_type = FLB_IO_TCP;
    }

    /* Prepare an upstream handler */
    upstream = flb_upstream_create(config,
                                   ins->host.name,
                                   ins->host.port,
                                   io_type,
                                   &ins->tls);
    if (!upstream) {
        flb_error("[out_es] cannot create Upstream context");
        flb_es_conf_destroy(ctx);
        return NULL;
    }

    /* Set manual Index and Type */
    ctx->u = upstream;
    if (f_index) {
        ctx->index = f_index->value;
    }
    else {
        tmp = flb_output_get_property("index", ins);
        if (!tmp) {
            ctx->index = flb_strdup(FLB_ES_DEFAULT_INDEX);
        }
        else {
            ctx->index = flb_strdup(tmp);
        }
    }

    if (f_type) {
        ctx->type = f_type->value;
    }
    else {
        tmp = flb_output_get_property("type", ins);
        if (!tmp) {
            ctx->type = flb_strdup(FLB_ES_DEFAULT_TYPE);
        }
        else {
            ctx->type = flb_strdup(tmp);
        }
    }

    /*
     * Logstash compatibility options
     * ==============================
     */

    /* Logstash_Format */
    tmp = flb_output_get_property("logstash_format", ins);
    if (tmp) {
        ctx->logstash_format = bool_value(tmp);
    }
    else {
        ctx->logstash_format = FLB_FALSE;
    }

    /* Logstash_Prefix */
    tmp = flb_output_get_property("logstash_prefix", ins);
    if (tmp) {
        ctx->logstash_prefix = flb_strdup(tmp);
        ctx->logstash_prefix_len = strlen(tmp);
    }
    else if (ctx->logstash_format == FLB_TRUE) {
        ctx->logstash_prefix = flb_strdup(FLB_ES_DEFAULT_PREFIX);
        ctx->logstash_prefix_len = sizeof(FLB_ES_DEFAULT_PREFIX) - 1;
    }

    /* Logstash_DateFormat */
    tmp = flb_output_get_property("logstash_dateformat", ins);
    if (tmp) {
        ctx->logstash_dateformat = flb_strdup(tmp);
        ctx->logstash_dateformat_len = strlen(tmp);
    }
    else if (ctx->logstash_format == FLB_TRUE) {
        ctx->logstash_dateformat = flb_strdup(FLB_ES_DEFAULT_TIME_FMT);
        ctx->logstash_dateformat_len = sizeof(FLB_ES_DEFAULT_TIME_FMT) - 1;
    }

    /* Time Key */
    tmp = flb_output_get_property("time_key", ins);
    if (tmp) {
        ctx->time_key = flb_strdup(tmp);
        ctx->time_key_len = strlen(tmp);
    }
    else {
        ctx->time_key = flb_strdup(FLB_ES_DEFAULT_TIME_KEY);
        ctx->time_key_len = sizeof(FLB_ES_DEFAULT_TIME_KEY) - 1;
    }

    /* Time Key Format */
    tmp = flb_output_get_property("time_key_format", ins);
    if (tmp) {
        ctx->time_key_format = flb_strdup(tmp);
        ctx->time_key_format_len = strlen(tmp);
    }
    else {
        ctx->time_key_format = flb_strdup(FLB_ES_DEFAULT_TIME_KEYF);
        ctx->time_key_format_len = sizeof(FLB_ES_DEFAULT_TIME_KEYF) - 1;
    }

    return ctx;
}

int flb_es_conf_destroy(struct flb_elasticsearch *ctx)
{
    flb_free(ctx->index);
    flb_free(ctx->type);

    flb_free(ctx->logstash_prefix);
    flb_free(ctx->logstash_dateformat);
    flb_free(ctx->time_key);
    flb_free(ctx->time_key_format);

    flb_upstream_destroy(ctx->u);
    flb_free(ctx);

    return 0;
}
